/*
   SPMmorph.c, written by John Villarrubia, National Institute of
   Standards and Technology, Gaithersburg, MD 20899, USA

   The algorithms presented below are intended to be used for research
   purposes only and bear no warranty, either express or implied.
   Please note that within the United States, copyright protection,
   under Section 105 of the United States Code, Title 17, is not
   available for any work of the United States Government and/or for
   any works conceived by United States Government employees under this
   software release. User acknowledges that Villarrubia's actual work
   is in the public domain and is not subject to copyright. However, if
   User utilizes the aforementioned government-created algorithms in a
   manner which substantailly alters User's work, User agrees to
   acknowledge this by reference to Villarrubia's papers, which are
   listed in the accompanying morph_lib.c file.

   This program is given as an example of how to use the routines in
   morph_lib.c. It is intended to help programmers who would like to
   include morph_lib routines in their own data analysis software.

   It can be used to perform actual data analysis on SPM images.
   However, the user interface is pretty primative for that purpose.

   This example program expects to read or write the tip or image in a
   format as follows. The file consists of a simple list of NxM+2
   integer (4 byte) entries, where N is the number of pixels/line and M
   is the number of lines. The entries are in this order, where x
   designates the pixel number within a line and y designates the line
   number:

	   N
	   M
	   data(x=0,y=0)
	   data(x=1,y=0)
	   .
	   .
	   .
	   data(x=N-1,y=0)
	   data(x=0,y=1)
	   data(x=1,y=1)
	   .
	   .
	   .
	   data(x=N-1,y=M-1)

   The file may be either an ASCII list of numbers separated by white
   space (spaces, tabs, newlines) or a binary list of numbers of type
   long.

   To compile and link, use cc SPMmorph.c morph_lib.c -o SPMmorph -O3
*/


#include <stdio.h>
#include <string.h>
#include <memory.h>

long **idilation();
long **ierosion();
long **icmap();
long **imatrix();
long **ireflect();
long **iopen();
long **readfile();
long getint();
void writefile();
void itip_estimate0();
long itip_estimate_point();
void itip_estimate();
long **allocmatrix();
void freematrix();

main() 
{
   long ans;
   char imagename[80],tipname[80],resultname[80],tempstring[80];
   long **image,**tip,**temp,**result;
   long impix,imlin,tippix,tiplin;	/* size of image and tip in pixels/line
					   and #lines */
   long resultpix,resultlin;		/* size of result */
   long oldpix,oldlin;
   long imflag=0,tipflag=0;	/* flags get set to 1 when these are read */
   long resultflag=0; 		/* =1 when there's a result */
   long xc,yc;			/* Coordinates of tip center (or apex) */
   long resultxc,resultyc;	/* xc and yc for result when result is a tip */
   long x,y;			/* coordinates of point in image for which
				   tip_estimate_pt is performed */
   long i,j;			/* loop counter */
   long thresh;			/* threshold value */

   strcpy(imagename,"NONE");
   strcpy(tipname,"NONE");
   strcpy(resultname,"NONE");

   for(;;) {
      /* Print menu of commands */
      printf("\
         1. Read surface or image (currently %s)\n\
         2. Read tip (currently %s)\n\
         3. Reflect image, tip, or result\n\
         4. Compute result = surface dilated by tip\n\
         5. Compute result = image eroded by tip\n\
         6. Compute result = tip_estimate0(image)\n\
         7. Compute result = tip_estimate_point(image)\n\
         8. Compute result = tip_estimate(image)\n\
         9. Compute result = certainty map (image,tip,rsurf)\n\
        10. Write result (currently %s)\n\
        11. Quit\n>> ",imagename,tipname,resultname);

      ans = getint();
      if (ans < 0) continue;	/* Error on command line */

      switch (ans) {

         case 1:		/* READ SURFACE OR IMAGE */
            temp = image; 
            oldpix = impix; 
            oldlin = imlin;
            strcpy(tempstring,imagename);
            printf("Input format: Choose one\n1 ASCII\n2 Binary\n>> ");
            ans = getint(); if (ans < 0) break;
            if (ans !=1 && ans !=2) {
               printf(
                  "You must respond by typing 1 for ASCII or 2 for Binary\n");
               break;
            }
            printf("Name of file containing image or surface: ");
            scanf("%s",imagename);
            image = readfile(imagename,ans,&impix,&imlin);
            if (image != NULL) { /*read succeeded. Set flag or free old array*/
               if (imflag==1) freematrix(temp,oldlin);
               else imflag=1;
            }
            else { /* read failed, restore previous image*/
               if (imflag==1) {image = temp; impix = oldpix; imlin = oldlin;}
               strcpy(imagename,tempstring);
            }
            break;

         case 2:		/* READ TIP */
            temp = tip; oldpix = tippix; oldlin = tiplin;
            strcpy(tempstring,tipname);
            printf("Input format: Choose one\n1 ASCII\n2 Binary\n>> ");
            ans = getint(); if (ans < 0) break;
            if (ans !=1 && ans !=2) {
               printf(
                  "You must respond by typing 1 for ASCII or 2 for Binary\n");
               break;
            }
            printf("Name of file containing tip: ");
            scanf("%s",tipname);
            tip = readfile(tipname,ans,&tippix,&tiplin);
            if (tip != NULL) { /*read succeeded. Set flag or free old array*/
               if (tipflag==1) freematrix(temp,oldlin);
               else tipflag=1;
               xc = tippix/2;	/* set xc & yc to defaults */
               yc = tiplin/2;
               printf("Default apex coordinates for this %ld x %ld array are",
                  tippix,tiplin);
               printf(" x=%ld, y=%ld\nUse these? (y or n) ",xc,yc);
               scanf("%s",tempstring);
               if (tempstring[0] != 'y' && tempstring[0] != 'Y') {
                  do {
                     printf("Enter coordinates in the ranges 0<=x<=%ld,",
                        tippix-1);
                     printf(" 0<=y<=%ld\n",tiplin-1);
                     printf("Enter x coordinate: ");
                     xc = getint();
                     printf("Enter y coordinate: ");
                     yc = getint();
                  } while(xc>=tippix || yc>=tiplin);
               }
            }
            else { /* read failed, restore previous image*/
               if (tipflag==1) {tip = temp; tippix = oldpix; tiplin = oldlin;}
               strcpy(tipname,tempstring);
            }
            break;

         case 3:		/* REFLECT */
            printf("   Choose one:\n\
   1 Replace image or surface by reflect(image) or reflect(surface)\n\
   2 Replace tip by reflect(tip)\n\
   3 Replace result by reflect(result)\n\
   >> ");
            ans = getint(); if (ans < 0) break;
            if (ans != 1 && ans != 2 && ans != 3) {
               printf("You must indicate either 1 or 2\n");
               break;
            }
            if (ans==1) {	/* Reflect image */
               if (imflag==0) {
                  printf("No image is currently defined.\n");
                  break;
               }
               temp = ireflect(image,impix,imlin);
               freematrix(image,imlin);
               image = temp;
            } else if (ans==2){	/* Reflect tip */
               if (tipflag==0) {
                  printf("No tip is currently defined.\n");
                  break;
               }
               temp = ireflect(tip,tippix,tiplin);
               freematrix(tip,tiplin);
               tip = temp;
               xc = tippix-1-xc;
               yc = tippix-1-yc;
            } else {		/* Reflect result */
               if (resultflag==0) {
                  printf("No result is currently defined.\n");
                  break;
               }
               temp = ireflect(result,resultpix,resultlin);
               freematrix(result,resultlin);
               result = temp;
               strcpy(resultname,"Reflection");
            }
            break;

         case 4:		/* DILATE */
            if (imflag==0) {
               printf("No surface is currently defined.\n");
               break;
            }
            if (tipflag==0) {
               printf("No tip is currently defined.\n");
               break;
            }
            if (resultflag) freematrix(result,resultlin);
            result = idilation(image,impix,imlin,tip,tippix,tiplin,xc,yc);
            resultpix = impix;
            resultlin = imlin;
            resultflag = 1;
            strcpy(resultname,"Dilation");
            break;

         case 5:		/* ERODE */
            if (imflag==0) {
               printf("No image is currently defined.\n");
               break;
            }
            if (tipflag==0) {
               printf("No tip is currently defined.\n");
               break;
            }
            if (resultflag) freematrix(result,resultlin);
            result = ierosion(image,impix,imlin,tip,tippix,tiplin,xc,yc);
            resultpix = impix;
            resultlin = imlin;
            resultflag = 1;
            strcpy(resultname,"Erosion");
            break;

         case 6:		/* TIP0 */
         case 7:		/* TIP_PT */
         case 8:		/* TIP */

            if (imflag==0) {
               printf("No image is currently defined.\n");
               break;
            }
            if (ans==7) {	/* Need PT coordinates for TIP_PT */
               printf("Calculation of tip estimate as derived from a single ");
               printf("image point.\n");
               printf("Enter coordinates within image for which you want this");
               printf(" calculation performed.\n");
               printf("Enter pixel # (x coordinate): ");
               x = getint();
               if (x<0 || x>impix-1) {
                  printf("x coordinate must be inside the image!"); break;
               }
               printf("Enter line # (y coordinate): ");
               y = getint();
               if (y<0 || y>imlin-1) {
                  printf("y coordinate must be inside the image!"); break;
               }
            }
            if (tipflag) {
               printf("(NB: The current tip is the one you read using ");
               printf("option #2 on the main menu,\nNOT the tip, if any, ");
               printf("resulting from your last calculation.) ");
               printf(
               "Use current tip as starting estimate for new one? (y or n) ");
            
               scanf("%s",tempstring);
               if (tempstring[0]=='y' || tempstring[0]=='Y') {
                  /* Copy tip to result */
                  if (resultflag==1) freematrix(result,resultlin);
                  resultlin = tiplin;
                  resultpix = tippix;
                  result = allocmatrix(resultlin,resultpix);
                  resultxc = xc;
                  resultyc = yc;
                  for (i=0;i<tiplin;i++) 
                     memcpy(result[i],tip[i],tippix*sizeof(long));
               }
            } 
            if (!tipflag || (tempstring[0]!='y' && tempstring[0]!='Y')) {
               printf("Starting estimate will be all zeros.\n");
               printf("Enter the array size for the resulting tip.\n");
               printf("     # lines? ");
               resultlin = getint();
               printf("     # pixels/line? ");
               resultpix = getint();
               result = allocmatrix(resultlin,resultpix);
               /* initialize array to zero */
	       for (j=0;j<resultlin;j++) for (i=0;i<resultpix;i++) 
                  result[j][i]=0;
               resultxc = resultpix/2; resultyc = resultlin/2;
               printf("Default lateral coordinates for tip apex are %ld,%ld. ",
                  resultxc,resultyc);
               printf("Use these? (y or n) ");
               scanf("%s",tempstring);
               if (tolower(tempstring[0])=='n') {
                  printf("Enter pixel # (x coordinate) of tip apex: ");
                  resultxc = getint();
                  if (resultxc<0 || resultxc>resultpix-1) {
                     printf("x coordinate must be inside the tip!"); break;
                  }
                  printf("Enter line # (y coordinate) of tip apex: ");
                  resultyc = getint();
                  if (resultyc<0 || resultyc>resultlin-1) {
                     printf("y coordinate must be inside the tip!"); break;
                  }
               }
            }
            resultflag = 1;
            printf("Enter threshold value to use: ");
            thresh = getint();
            printf("Beginning calculation.\n");
            switch (ans) {
               case 6:
                  itip_estimate0(image,impix,imlin,resultpix,resultlin,resultxc,
                     resultyc,result,thresh);
                  printf("Calculation of tip0 completed.\n");
                  strcpy(resultname,"recon_tip0");
                  break;
               case 7:
                  i=itip_estimate_point(x,y,image,impix,imlin,resultpix,
                     resultlin,resultxc,resultyc,result,thresh);
                  printf("Calculation of tip_pt completed.\n");
                  printf("Starting tip estimate was updated at %ld pixels.\n",i);
                  strcpy(resultname,"recon_tip_pt");
                  break;
               case 8:
                  itip_estimate(image,impix,imlin,resultpix,resultlin,resultxc,
                     resultyc,result,thresh);
                  printf("Calculation of tip completed.\n");
                  strcpy(resultname,"recon_tip");
                  break;
            }
            break;

         case 9:		/* COMPUTE CERTAINTY MAP */
            printf("Computation of Pingali-type certainty");
            printf(" map has these requirements.\n1. You must have loaded");
            printf(" an image. (Menu command #1)\n2. You must have loaded");
            printf(" a tip. (Menu command #2)\n3. The result buffer must");
            printf(" contain erosion(image,tip). (Menu command #5)\n");

            if (imflag==0 || tipflag==0 || resultflag==0) {
               if (imflag == 0) printf("Image is undefined.\n");
               if (tipflag == 0) printf("Tip is undefined.\n");
               if (resultflag == 0) printf("Result is undefined.\n");
               break;
            }

            printf("Continue? (y or n) ");
            scanf("%s",tempstring);
            if (tempstring[0]!='y' && tempstring[0]!='Y') break;

            temp = icmap(image,impix,imlin,tip,tippix,tiplin,result,xc,yc);
            if (temp != NULL) {		/* succeeded */
               freematrix(result,resultlin);
               strcpy(resultname,"Certainty map");
               resultpix = impix;
               resultlin = imlin;
               result = temp;
            }
            break;

         case 10:		/* WRITE */
            if (resultflag==0) {
               printf("No result currently exists.\n");
               break;
            }
            printf("Output format: Choose one\n1 ASCII\n2 Binary\n>> ");
            ans = getint(); if (ans < 0) break;
            if (ans !=1 && ans !=2) {
               printf(
                  "You must respond by typing 1 for ASCII or 2 for Binary\n");
               break;
            }
            printf("Name of file for output: ");
            scanf("%s",resultname);
            writefile(resultname,ans,resultpix,resultlin,result);
            break;

         case 11:		/* QUIT */
            exit(0);
            break;

         default:		/* none of the above */
            printf("%ld is not a valid response. Try again.\n",ans);
            break;

      } /* end switch */
   } /* end for */
}

/* 
   The function getint() obtains a long integer from the keyboard. The
   reason we don't use simply scanf("%ld",&ans) is that in that case if
   the user accidentally types a non-numeric character we're in big
   trouble. scanf() leaves characters inconsistent with the %ld format
   in the input buffer.  They'll still be there next time we try to
   read an integer, so any "invalid response" error message simply gets
   repeated over and over again.  So instead, we read the whole input
   line and discard it if there are invalid characters.

   getint() returns an integer if it gets one, -1 if error.
*/
   
long getint()		/* Gets an integer from the user */
{
   char *chptr,response[100],*digptr,digits[100];
   long ans;

   digptr = digits;
   chptr=fgets(response,99,stdin);	 /* read whole line */
   /* If 1st character was new line, get more */
   if (*chptr == '\n') chptr=fgets(response,99,stdin); 
   while (isspace(*chptr)) chptr++;	/* eat white space if any */
   if (isdigit(*chptr)) {
      while(isdigit(*chptr)) *digptr++ = *chptr++;
      *digptr = '\0';	/* terminate string */
      sscanf(digits,"%ld",&ans);	/* read integer */
      while (isspace(*chptr)) chptr++;	/* eat any more white space */
      if (*chptr != '\0') {	/* next character better be end-of-string */
         printf("Extra characters on line. You must enter a single integer.\n");
         return(-1);
      }
      return(ans);
   } 
   printf("Non-numeric characters on line. You must enter a single positive");
   printf(" integer.\n");
   return(-1);
}

long **readfile(filename,binflag,pixels,lines)
   char *filename;
   long binflag;		/* =1 for ASCII, =2 for binary */
   long *pixels,*lines;
{
   FILE *fileptr;
   long **array;
   long i,j;
   
   if (binflag==1) {
      fileptr = fopen(filename,"r");
      if (fileptr==NULL) {
         printf("File not found\n");
         return(NULL);
      }
      /* read pixels and lines */
      fscanf(fileptr,"%ld",pixels);
      fscanf(fileptr,"%ld",lines);
      /* allocate array and read data into it */
      array = allocmatrix(*lines,*pixels);
      for (j=0;j<*lines;j++) for (i=0;i<*pixels;i++) 
         fscanf(fileptr,"%ld",array[j]+i);
      fclose(fileptr);
      return(array);
   } else if (binflag==2) {	/* Binary file */
      fileptr = fopen(filename,"rb");
      if (fileptr==NULL) {
         printf("File not found\n");
         return(NULL);
      }
      /* read pixels and lines */
      fread(pixels,sizeof(long),1,fileptr);
      fread(lines,sizeof(long),1,fileptr);
      /* allocate array and read data into it */
      array = allocmatrix(*lines,*pixels);
      for (j=0;j<*lines;j++) fread(array[j],sizeof(long),*pixels,fileptr);
      fclose(fileptr);
      return(array);
   } else {		/* invalid format */
      printf("Invalid format specification\n");
      return(NULL);
   }
}

void writefile(filename,binflag,pixels,lines,array)
   char *filename;
   long binflag;		/* =1 for ASCII, =2 for binary */
   long pixels,lines;
   long **array;
{
   FILE *fileptr;
   long i,j;
   
   if (binflag==1) {
      fileptr = fopen(filename,"w");
      if (fileptr==NULL) {
         printf("Can't open file for output. Nothing written.\n");
         return;
      }
      /* write pixels and lines */
      fprintf(fileptr,"%ld\n",pixels);
      fprintf(fileptr,"%ld\n",lines);
      /* write the array */
      for (j=0;j<lines;j++) for (i=0;i<pixels;i++) 
         fprintf(fileptr,"%ld\n",array[j][i]);
      fclose(fileptr);
      return;
   } else if (binflag==2) {	/* Binary file */
      fileptr = fopen(filename,"wb");
      if (fileptr==NULL) {
         printf("Can't open file for output. Nothing written.\n");
         return;
      }
      /* write pixels and lines */
      fwrite(&pixels,sizeof(long),1,fileptr);
      fwrite(&lines,sizeof(long),1,fileptr);
      /* write the array */
      for (j=0;j<lines;j++) fwrite(array[j],sizeof(long),pixels,fileptr);
      fclose(fileptr);
      return;
   } else {		/* invalid format */
      printf("Invalid format specification\n");
      return;
   }
}
