/* 
   Library of mathematical morphology and tip estimation routines
   written by John Villarrubia, National Institute of Standards and
   Technology, an agency of the U.S. Department of Commerce,
   Gaithersburg, MD 20899, USA.

   The algorithms presented below are intended to be used for research
   purposes only and bear no warranty, either express or implied.
   Please note that within the United States, copyright protection,
   under Section 105 of the United States Code, Title 17, is not
   available for any work of the United States Government and/or for
   any works conceived by United States Government employees under this
   software release. User acknowledges that Villarrubia's actual work
   is in the public domain and is not subject to copyright. However, if
   User utilizes the aforementioned government-created algorithms in a
   manner which substantially alters User's work, User agrees to
   acknowledge this by reference to Villarrubia's papers, which are
   listed below.

   The routines include:

UTILITY ROUTINES

	long **allocmatrix()	Allocates memory and returns a pointer
				to the kind of 2-d array required to
				represent images, surfaces, and tips by
				the rest of the routines.
	void freematrix()	Frees an array of the above type.
	long **ireflect()	By convention the geometry for SPM is 
				pictured with a downward-pointing tip
				above a surface. The "tip" used as
				input or calculated by all of the
				following routines is actually the
				upward-pointing REFLECTION of the
				actual tip through the origin. The
				ireflect() routine is provided as a
				utility to convert back and forth
				between the true geometry and the
				reflected one.

BASIC MATHEMATICAL MORPHOLOGY AND CERTAINTY MAP ROUTINES

	long **idilation()	Computes dilation. Used for simulating
				imaging given a known surface and tip.
	long **ierosion()	Computes erosion. Used for surface
				reconstruction given an image and a tip
				estimate, or tip reconstruction given an
				image and a known reference surface.
	long **iopen()		Computes the opening, which is an erosion
				followed by a dilation. Used by the 
				blind tip estimation routine, itip_estimate().
	long **icmap()		Computes a Pingali-type certainty map given
				an image, tip, and associated reconstructed 
				surface.
BLIND TIP ESTIMATION ROUTINES

	long itip_estimate_point()	Computes the largest tip with a 
				given "footprint" consistent with an
				image at a given point. Used by
				itip_estimate() and itip_estimate0()
	void itip_estimate()	Performs a full blind reconstruction from an
				image by applying itip_estimate_point to all
				pixels within the image, and then iterating
				until convergence.
	void itip_estimate0()	Performs a partial blind reconstruction from
				an image by applying
				itip_estimate_point() to a subset of
				points within the image. The points in
				this subset should be chosen from among
				those parts on the image which contain
				the most information about the tip, for
				example, isolated local maxima or
				points of high curvature. This is
				faster than the full itip_estimate(),
				and may usefully provide a starting
				estimate to the full blind
				reconstruction routine.
	long useit()		As routine which establishes
				the selection criterion for the subset
				of points used in itip_estimate0().
				Given an image and coordinates, useit()
				returns a 1 if the point at those
				coordinates fits the criterion and 0 if
				it does not. As supplied, the criterion
				employed is that the point be the
				maximum within some neighborhood. The
				user can change the criterion by
				rewriting this routine.


REFERENCES

   For the development of the theory underlying these routines, the
   following references may prove useful:

   1. "Algorithms for scanned probe microscope image simulation,
   surface reconstruction, and tip estimation," J.S. Villarrubia, J.
   Res. Natl. Inst. Stand. & Technol. 102(4), 425 (1997).
   --Essentially the user's guide to this software, it includes
   theoretical background and a listing of the code with notes for users
   and programmers.

   2. "Morphological estimation of tip geometry for scanned probe
   microscopy," J.S. Villarrubia, Surf. Sci. 321, 287 (1994).
   --The original paper describing blind tip estimation.

   3. "Restoration of scanning probe microscope images," G.S. Pingali and
   R. Jain, Proceedings IEEE Workshop on Applications of Computer
   Vision, 282 (1992).
   --An early rigorous connection between mathematical morphology and
   SPM imaging, contains discussion relevant to image simulation
   (dilation), surface reconstruction (in terms of Minkowski
   subtraction instead of erosion, however), and certainty maps.

   4. "Scanned probe microscope tip characterization without calibrated
   tip characterizers," J.S. Villarrubia, J. Vac. Sci. Technol. B 14,
   1518 (1996).
   --A demonstration of some of the potential uses of blind tip
   estimation, and a comparison to tip estimation by imaging a known
   reference artifact.

   5. "Blind reconstruction of scanning probe image data," P.M.
   Williams, K.M.  Shakesheff, M.C. Davies, D.E. Jackson, C.J. Roberts,
   J. Vac. Sci. Technol.  B 14, 1557 (1996).
   --This group independently developed essentially the same blind
   reconstruction method which I described in [2] above and embodied in
   itip_estimate(), below. This paper is their description of the
   method. Whereas [2] is more formal and mathematical (equations,
   proofs, etc.) this paper is more algorithm-oriented and descriptive,
   a style which complements [2] and may be more appealing to some
   readers.

ADDITIONAL NOTES ON USAGE

   This file is distributed with a companion "main" program in the
   file SPMmorph.c. SPMmorph provides a rather primitive interface
   to this library, which may be useful for simple analysis and as
   an example of the use of these routines for programmers who wish
   to incorporate them into more user-friendly software. To compile
   and link with SPMmorph, I use

	cc SPMmorph.c morph_lib.c -o SPMmorph -O3

   (The -O3 is an optimization switch which may be omitted.)

   Several additional modules of this library are intended to be called
   directly from PV-WAVE using its LINKNLOAD routine. Their names are
   the same as the above routines, but without the leading "i" (e.g.
   dilation instead of idilation). Each of these routines essentially
   provides a "wrapper" or interface, which converts data as passed
   from PV-WAVE to the form required by the morphology routines,
   computes the result, and then passes the result back to PV-WAVE. For
   more specifics, see the PV-WAVE documentation. To compile for PV-WAVE
   on my SUN workstation, I use:

		cc -c -pic morph_lib.c -O3
		ld -o morph_lib.so -assert pure-text morph_lib.o

    My understanding is that PV-WAVE 6.0 for PC's now allows similar
    dynamic linking, though I expect compilation requirements may be 
    system dependent.
*/


#include <stdio.h>
#include <memory.h>

#define MIN(a,b) ((a)<(b)?a:b)
#define MAX(a,b) ((a)>(b)?a:b)

long dilation(),erosion(),tip_estimate(),reflect(),tip_estimate_point(),
    tip_estimate0(),cmap();
long **ireflect();
long **idilation();
long **ierosion();
long **icmap();
long **iopen();
long **allocmatrix();
long itip_estimate_iter();
long itip_estimate_point();
long useit();
void itip_estimate();
void itip_estimate0();
void freematrix();


/* The following routines serve as an interface between PV-WAVE and the
   relect, dilation, erosion, and tip estimation routine(s) below. They
   can be called directly from PV-WAVE using the LINKNLOAD utility.
   They receive addresses of the PV-WAVE variables, convert these to a
   form accessible to the ireflect, idilation, ierosion, and
   itip_estimation (i*) routines, use the i* routines to do the actual
   operations, and then return.
*/
long reflect(argc,argv)
   int argc;
   char *argv[];
{
   long *image,*xsiz,*ysiz,*result;
   long **im,**ref;
   long i,j;

   if (argc != 4) {
      printf("reflect routine called with wrong number (%d) of parameters\n",
         argc);
      return(1);
   }

   image = ((long **)argv)[0];
   xsiz = ((long **)argv)[1];
   ysiz = ((long **)argv)[2];
   result = ((long **)argv)[3];

   im = (long **)malloc(*ysiz*sizeof(long*));
   if (im==NULL) {
      printf("Memory allocation failure #1 in reflect.\n");
      return(1);
   }

   
   for (i=0;i<*ysiz;i++) im[i]=image+*xsiz*i;

   /* Compute the reflection */
   ref = ireflect(im,*xsiz,*ysiz);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   for (j=0;j<*ysiz;j++) memcpy(result+j*(*xsiz),ref[j],(*xsiz)*sizeof(long));

   free(im); freematrix(ref,*ysiz);
   return(0);
}


long dilation(argc,argv)
   int argc;
   char *argv[];
{
   long *surface,*xsiz,*ysiz,*tip,*tip_xsiz,*tip_ysiz,*xc,*yc,*result;
   long **surf,**tp,**dil;
   long i,j;

   if (argc != 9) {
      printf("dilation routine called with wrong number (%d) of parameters\n",
         argc);
      return(1);
   }

   surface = ((long **)argv)[0];
   xsiz = ((long **)argv)[1];
   ysiz = ((long **)argv)[2];
   tip = ((long **)argv)[3];
   tip_xsiz = ((long **)argv)[4];
   tip_ysiz = ((long **)argv)[5];
   xc = ((long **)argv)[6];
   yc = ((long **)argv)[7];
   result = ((long **)argv)[8];

   surf = (long **)malloc(*ysiz*sizeof(long*));
   if (surf==NULL) {
      printf("Memory allocation failure #1 in dilation.\n");
      return(1);
   }

   tp = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp==NULL) {
      printf("Memory allocation failure #2 in dilation.\n");
      free(surf);
      return(1);
   }
   
   for (i=0;i<*ysiz;i++) surf[i]=surface+*xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp[i] = tip+*tip_xsiz*i;

   /* Compute the dilation */
   dil = idilation(surf,*xsiz,*ysiz,tp,*tip_xsiz,*tip_ysiz,*xc,*yc);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   /* I could save having to do the following copy if I make
      idilation,ierosion,iopen return their result in the argument list
      rather than as a function. However, the copy then needs to be
      done inside of the dilation or erosion routine, hence no time
      savings. */

   for (j=0;j<*ysiz;j++) memcpy(result+j*(*xsiz),dil[j],(*xsiz)*sizeof(long));

   free(surf); free(tp); freematrix(dil,*ysiz);
   return(0);
}

long erosion(argc,argv)
   int argc;
   char *argv[];
{
   long *image,*xsiz,*ysiz,*tip,*tip_xsiz,*tip_ysiz,*xc,*yc,*result;
   long **img,**tp,**eros;
   long i,j;

   if (argc != 9) {
      printf("erosion routine called with wrong number (%d) of parameters\n",
         argc);
      return(1);
   }

   image = ((long **)argv)[0];
   xsiz = ((long **)argv)[1];
   ysiz = ((long **)argv)[2];
   tip = ((long **)argv)[3];
   tip_xsiz = ((long **)argv)[4];
   tip_ysiz = ((long **)argv)[5];
   xc = ((long **)argv)[6];
   yc = ((long **)argv)[7];
   result = ((long **)argv)[8];

   img = (long **)malloc(*ysiz*sizeof(long*));
   if (img==NULL) {
      printf("Memory allocation failure #1 in erosion.\n");
      return(1);
   }

   tp = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp==NULL) {
      printf("Memory allocation failure #2 in erosion.\n");
      free(img);
      return(1);
   }
   
   for (i=0;i<*ysiz;i++) img[i]=image+*xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp[i] = tip+*tip_xsiz*i;

   /* Compute the erosion */
   eros = ierosion(img,*xsiz,*ysiz,tp,*tip_xsiz,*tip_ysiz,*xc,*yc);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space. */

   for (j=0;j<*ysiz;j++) memcpy(result+j*(*xsiz),eros[j],(*xsiz)*sizeof(long));

   free(img); free(tp); freematrix(eros,*ysiz);
   return(0);
}

long cmap(argc,argv)
   int argc;
   char *argv[];
{
   long *image,*xsiz,*ysiz,*tip,*tip_xsiz,*tip_ysiz,*rsurf,*xc,*yc,*map;
   long **im,**tp,**rs,**cmap;
   long i,j;

   if (argc != 10) {
      printf("cmap routine called with wrong number (%d) of parameters\n",argc);
      return(1);
   }

   image = ((long **)argv)[0];
   xsiz = ((long **)argv)[1];
   ysiz = ((long **)argv)[2];
   tip = ((long **)argv)[3];
   tip_xsiz = ((long **)argv)[4];
   tip_ysiz = ((long **)argv)[5];
   rsurf = ((long **)argv)[6];
   xc = ((long **)argv)[7];
   yc = ((long **)argv)[8];
   map = ((long **)argv)[9];


   im = (long **)malloc(*ysiz*sizeof(long*));
   if (im==NULL) {
      printf("Memory allocation failure #1 in cmap.\n");
      return(1);
   }

   tp = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp==NULL) {
      printf("Memory allocation failure #2 in cmap.\n");
      free(im);
      return(1);
   }

   rs = (long **)malloc(*ysiz*sizeof(long*));
   if (rs==NULL) {
      printf("Memory allocation failure #3 in cmap.\n");
      free(im); free(tp);
      return(1);
   }
   
   for (i=0;i<*ysiz;i++) im[i]=image+*xsiz*i;
   for (i=0;i<*ysiz;i++) rs[i]=rsurf+*xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp[i] = tip+*tip_xsiz*i;

   /* Compute the cmap */
   cmap = icmap(im,*xsiz,*ysiz,tp,*tip_xsiz,*tip_ysiz,rs,*xc,*yc);
   if (cmap == NULL) return(1);		/* Error occurred */
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   /* I could save having to do the following copy if I make
      idilation,ierosion,iopen return their result in the argument list
      rather than as a function. However, the copy then needs to be
      done inside of the dilation or erosion routine, hence no time
      savings. */

   for (j=0;j<*ysiz;j++) memcpy(map+j*(*xsiz),cmap[j],(*xsiz)*sizeof(long));

   free(im); free(tp); free(rs); freematrix(cmap,*ysiz);
   return(0);
}


long tip_estimate(argc,argv)
   int argc;
   char *argv[];
{
   long *image,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,
       *xc,*yc,*tip0,*thresh,*result;
   long **img,**tp0;
   long i,j;

   if (argc != 10) {
   printf("tip_estimate routine called with wrong number (%d) of parameters\n",
      argc);
      return(1);
   }

   image = ((long **)argv)[0];
   im_xsiz = ((long **)argv)[1];
   im_ysiz = ((long **)argv)[2];
   tip_xsiz = ((long **)argv)[3];
   tip_ysiz = ((long **)argv)[4];
   xc = ((long **)argv)[5];
   yc = ((long **)argv)[6];
   tip0 = ((long **)argv)[7];
   thresh = ((long **)argv)[8];
   result = ((long **)argv)[9];

   img = (long **)malloc(*im_ysiz*sizeof(long*));
   if (img==NULL) {
      printf("Memory allocation failure #1 in tip_estimate.\n");
      return(1);
   }

   tp0 = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp0==NULL) {
      printf("Memory allocation failure #2 in tip_estimate.\n");
      free(img);
      return(1);
   }
   
   for (i=0;i<*im_ysiz;i++) img[i]=image+*im_xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp0[i] = tip0+*tip_xsiz*i;

   /* Compute the tip_estimate */
   itip_estimate(img,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,*xc,*yc,tp0,*thresh);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   for (j=0;j<*tip_ysiz;j++) 
      memcpy(result+j*(*tip_xsiz),tp0[j],(*tip_xsiz)*sizeof(long));

   free(img); free(tp0);

   return(0);
}

long tip_estimate0(argc,argv)
   int argc;
   char *argv[];
{
   long *image,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,
       *xc,*yc,*tip0,*thresh,*result;
   long **img,**tp0;
   long i,j;

   if (argc != 10) {
   printf("tip_estimate0 routine called with wrong number (%d) of parameters\n",
      argc);
      return(1);
   }

   image = ((long **)argv)[0];
   im_xsiz = ((long **)argv)[1];
   im_ysiz = ((long **)argv)[2];
   tip_xsiz = ((long **)argv)[3];
   tip_ysiz = ((long **)argv)[4];
   xc = ((long **)argv)[5];
   yc = ((long **)argv)[6];
   tip0 = ((long **)argv)[7];
   thresh = ((long **)argv)[8];
   result = ((long **)argv)[9];

   img = (long **)malloc(*im_ysiz*sizeof(long*));
   if (img==NULL) {
      printf("Memory allocation failure #1 in tip_estimate.\n");
      return(1);
   }

   tp0 = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp0==NULL) {
      printf("Memory allocation failure #2 in tip_estimate.\n");
      free(img);
      return(1);
   }
   
   for (i=0;i<*im_ysiz;i++) img[i]=image+*im_xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp0[i] = tip0+*tip_xsiz*i;

   /* Compute the tip_estimate */
   itip_estimate0(img,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,
      *xc,*yc,tp0,*thresh);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   for (j=0;j<*tip_ysiz;j++) 
      memcpy(result+j*(*tip_xsiz),tp0[j],(*tip_xsiz)*sizeof(long));

   free(img); free(tp0);

   return(0);
}

long tip_estimate_point(argc,argv)
   int argc;
   char *argv[];
{
   long *ix,*iy,*image,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,
       *xc,*yc,*tip0,*thresh,*result;
   long **img,**tp0;
   long i,j;

   if (argc != 12) {
   printf("tip_estimate_point called with wrong number (%d) of parameters\n",
      argc);
      return(1);
   }

   ix = ((long **)argv)[0];
   iy = ((long **)argv)[1];
   image = ((long **)argv)[2];
   im_xsiz = ((long **)argv)[3];
   im_ysiz = ((long **)argv)[4];
   tip_xsiz = ((long **)argv)[5];
   tip_ysiz = ((long **)argv)[6];
   xc = ((long **)argv)[7];
   yc = ((long **)argv)[8];
   tip0 = ((long **)argv)[9];
   thresh = ((long **)argv)[10];
   result = ((long **)argv)[11];

   img = (long **)malloc(*im_ysiz*sizeof(long*));
   if (img==NULL) {
      printf("Memory allocation failure #1 in tip_estimate.\n");
      return(1);
   }

   tp0 = (long **)malloc(*tip_ysiz*sizeof(long*));
   if (tp0==NULL) {
      printf("Memory allocation failure #2 in tip_estimate.\n");
      free(img);
      return(1);
   }
   
   for (i=0;i<*im_ysiz;i++) img[i]=image+*im_xsiz*i;
   for (i=0;i<*tip_ysiz;i++) tp0[i] = tip0+*tip_xsiz*i;

   /* Compute the tip_estimate */
   itip_estimate_point(*ix,*iy,img,*im_xsiz,*im_ysiz,*tip_xsiz,*tip_ysiz,
      *xc,*yc,tp0,*thresh);
   
   /* Copy the result into the PV-WAVE variable. Then free intermediate
      variable space */

   for (j=0;j<*tip_ysiz;j++) 
      memcpy(result+j*(*tip_xsiz),tp0[j],(*tip_xsiz)*sizeof(long));

   free(img); free(tp0);

   return(0);
}


/* The following routines allow allocation and freeing of matrices */

long **allocmatrix(ysiz,xsiz)
long ysiz,xsiz;
/* 
   Allocates an integer matrix of dimension [ysiz][xsiz] using an array
   of pointers to rows. ysiz is the number of rows. xsiz is the number
   of columns.
*/
{
   long **mptr,		/* points to allocated matrix */
       i;		/* counter */

   /* Allocate pointers to rows */
   mptr = (long **)malloc(ysiz*sizeof(long*));
   if (mptr == NULL) {
      printf("Error: Allocation of mptr failed in allocmatrix\n");
      exit(1);
   }
   

   /* Allocate rows */
   for (i=0;i<ysiz;i++) {
      mptr[i]=(long *)malloc(xsiz*sizeof(long));
      if (mptr[i] == NULL) {
         printf("Error: Allocation of mptr[%ld] failed in allocmatrix\n",i);
         exit(1);
      }
   }  
   
   /* Done. Return result. */
   return mptr;
}

void freematrix(mptr,ysiz)
   long **mptr;
   long ysiz;
/* Frees memory allocated with allocmatrix */
{
   long i;
   for (i=0;i<ysiz;i++) free(mptr[i]);
   free(mptr);
}

/* The following routine performs reflection of integer arrays. The integers
   used are standard C integers, which PV-WAVE would call long integers. */

long **ireflect(surface,surf_xsiz,surf_ysiz)
   long **surface,   /* surface is a [0...surf_ysiz-1][0...surf_xsiz-1] array */
       surf_xsiz,surf_ysiz;                   /* size of surface */
{
   long **result;
   long i,j;               /* index */

   /* create output array of appropriate size */
   result = allocmatrix(surf_ysiz,surf_xsiz);

   for (j=0;j<surf_ysiz;j++) {   /* Loop over all points in output array */
      for (i=0;i<surf_xsiz;i++) {
	 result[j][i] = -surface[surf_ysiz-1-j][surf_xsiz-1-i];
      }
   }
   return(result);
}


/* The following routine performs dilation on integer arrays. The integers
   used are standard C integers (for my compiler 4-byte), which PV-WAVE
   would call long integers. */

long **idilation(surface,surf_xsiz,surf_ysiz,
	      tip,tip_xsiz,tip_ysiz,xc,yc)
   long **surface,    /* surface is a [0...surf_ysiz-1][0...surf_xsiz-1] array */
       surf_xsiz,surf_ysiz,                   /* size of surface */
       **tip,        /* tip is a [0...tip_ysiz-1][0...tip_xsiz-1] array */
       tip_xsiz,tip_ysiz,                   /* size of tip */
       xc,yc;                   /* center coordinates of tip */
{
   long **result;
   long i,j,px,py;               /* index */
   long max;
   long pxmin,pxmax,pymin,pymax; /* range of indices into tip */
   long temp;

   /* create output array of appropriate size */
   result = allocmatrix(surf_ysiz,surf_xsiz);

   for (j=0;j<surf_ysiz;j++) {   /* Loop over all points in output array */
      /* Compute allowed range of py. This may be different from
	 the full range of the tip due to edge overlaps. */
      pymin = MAX(j-surf_ysiz+1,-yc);
      pymax = MIN(tip_ysiz-yc-1,j);
      for (i=0;i<surf_xsiz;i++) {
	 /* Compute allowed range of px. This may be different from
	    the full range of the tip due to edge overlaps. */
	 pxmin = MAX(i-surf_xsiz+1,-xc);
	 pxmax = MIN(tip_xsiz-xc-1,i);
	 max = surface[j-pymin][i-pxmin]+tip[pymin+yc][pxmin+xc];
	 for (px=pxmin;px<=pxmax;px++) {        /* Loop over points in tip */
	    for (py=pymin;py<=pymax;py++) {
	       temp = surface[j-py][i-px]+tip[py+yc][px+xc];
	       max = MAX(temp,max);
	    }
	 }
	 result[j][i] = max;
      }
   }
   return(result);
}

long **ierosion(image,im_xsiz,im_ysiz,
	      tip,tip_xsiz,tip_ysiz,
	      xc,yc)
   long **image,    /* image is a [0...im_ysiz-1][0...im_xsiz-1] array */
       im_xsiz,im_ysiz,                   /* size of image */
       **tip,        /* tip is a [0...tip_ysiz-1][0...tip_xsiz-1] array */
       tip_xsiz,tip_ysiz,                   /* size of tip */
       xc,yc;                   /* center coordinates of tip */
{
   long **result;
   long i,j,px,py;               /* index */
   long min;
   long pxmin,pxmax,pymin,pymax; /* range of indices into tip */
   long temp;

   /* create output array of appropriate size */
   result = allocmatrix(im_ysiz,im_xsiz);

   for (j=0;j<im_ysiz;j++) {   /* Loop over all points in output array */
      /* Compute allowed range of py. This may be different from
	 the full range of the tip due to edge overlaps. */
      pymin = MAX(-j,-yc);
      pymax = MIN(tip_ysiz-yc,im_ysiz-j)-1;
      for (i=0;i<im_xsiz;i++) {
	 /* Compute allowed range of px. This may be different from
	    the full range of the tip due to edge overlaps. */
	 pxmin = MAX(-xc,-i);
	 pxmax = MIN(tip_xsiz-xc,im_xsiz-i)-1;
	 min = image[j+pymin][i+pxmin]-tip[pymin+yc][pxmin+xc];
	 for (px=pxmin;px<=pxmax;px++) {        /* Loop over points in tip */
	    for (py=pymin;py<=pymax;py++) {
	       temp = image[j+py][i+px]-tip[py+yc][px+xc];
	       min = MIN(temp,min);
	    }
	 }
	 result[j][i] = min;
      }
   }
   return(result);
}

long **icmap(image,im_xsiz,im_ysiz,
	      tip,tip_xsiz,tip_ysiz,
              rsurf,
	      xc,yc)
   long **image,    /* image is a im_ysiz x im_xsiz array */
       im_xsiz,im_ysiz,                   /* size of image */
       **tip,        /* tip is a tip_ysiz x tip_xsiz array */
       tip_xsiz,tip_ysiz,                   /* size of tip */
       **rsurf,		/* rsurf is a im_ysiz x im_xsiz array */
       xc,yc;                   /* center coordinates of tip */
{
   long **cmap;
   long imx,imy,tpx,tpy;               /* index */
   long tpxmin,tpxmax,tpymin,tpymax;
   long count;
   long rxc,ryc;			/* center coordinates of reflected tip */
   long x,y;

   rxc = tip_xsiz-1-xc;
   ryc = tip_ysiz-1-yc;

   /* create output array of appropriate size */
   cmap = allocmatrix(im_ysiz,im_xsiz);
   for (imy=0;imy<im_ysiz;imy++) for (imx=0;imx<im_xsiz;imx++) cmap[imy][imx]=0;

   /*
       Loop over all pixels in the interior of the image. We skip
       pixels near the edge. Since it is possible there are unseen
       touches over the edge, we must conservatively leave these cmap
       entries at 0.
   */
   for (imy=ryc;imy<=im_ysiz+ryc-tip_ysiz;imy++) {   
      for (imx=rxc;imx<=im_xsiz+rxc-tip_xsiz;imx++) {
         tpxmin = MAX(0,rxc-imx);
         tpxmax = MIN(tip_xsiz-1,im_xsiz-1+rxc-imx);
         tpymin = MAX(0,ryc-imy);
         tpymax = MIN(tip_ysiz-1,im_ysiz-1+ryc-imy);
         count = 0;
         for (tpy=tpymin;tpy<=tpymax&&count<2;tpy++) {
            for (tpx=tpxmin;tpx<=tpxmax&&count<2;tpx++) {
               if (image[imy][imx]-tip[tip_ysiz-1-tpy][tip_xsiz-1-tpx] == 
                  rsurf[tpy+imy-ryc][tpx+imx-rxc]) {
                  count++;	/* increment count */
                  x = tpx+imx-rxc;	/* remember coordinates */
                  y = tpy+imy-ryc;
               }
            }
         }
         if (count==1) cmap[y][x] = 1;	/* 1 contact = good recon */
      }
   }
   return(cmap);
}

long **iopen(image,im_xsiz,im_ysiz,tip,tip_xsiz,tip_ysiz)
   long **image,    /* image is a im_xsiz by im_ysiz array */
       im_xsiz,im_ysiz,     /* size of image */
       **tip,        /* tip is a tip_ysiz by tip_xsiz array */
       tip_xsiz,tip_ysiz;                   /* size of tip */
{
   long **result,**eros;

   eros = ierosion(image,im_xsiz,im_ysiz,tip,tip_xsiz,tip_ysiz,
      tip_xsiz/2,tip_ysiz/2);
   result = idilation(eros,im_xsiz,im_ysiz,tip,tip_xsiz,tip_ysiz,
      tip_xsiz/2,tip_ysiz/2);
   freematrix(eros,im_ysiz);/* free intermediate result */
   return(result);
}

/* The following routine estimates tip size by calling tip_estimate_iter
   until it converges. */

void itip_estimate(image,im_xsiz,im_ysiz,tip_xsiz,tip_ysiz,xc,yc,tip0,thresh)
   long **image,            /* image array */
       im_xsiz,im_ysiz,    /* x,y size of image array */   
       tip_xsiz,tip_ysiz,  /* x,y, size of tip to be determined */
       xc,yc,              /* center coordinates of the tip */
       **tip0,   /* starting estimate. Must be tip_xsiz by tip_ysiz in size */
       thresh;             /* threshold value to use */
{
   long iter=0;
   long count=1;

   while (count) {
      iter++;
      count = itip_estimate_iter(image,im_xsiz,im_ysiz,
         tip_xsiz,tip_ysiz,xc,yc,tip0,thresh);
      printf ("Finished iteration #%ld. ",iter);
      printf ("%ld image locations produced refinement.\n",count);
   }
}

/* The following routine performs one iteration of the tip estimation
   recursion algorithm. The threshold parameter is used as follows:  A
   new estimate of the tip height at a given pixel is computed
   according to the formula given by Eqn 13 (or 16) of Surf. Sci. 321
   (1994) 287. This value is then augmented by thresh. If the old
   estimate is less than the augmented new estimate, no action is taken
   Otherwise, the value is changed to agree with the augmented new
   estimate. Thus, if thresh=0, this routine implements Eqn 13
   directly. Larger values of threshold give greater noise immunity at
   the cost of degrading the estimate.

   In this version of the routine, the values of the revised estimate
   replace those in tip0, and the number of pixels where the value was
   changed is returned.
   */

#define MINUS_INF -2147483648L           /* smallest integer */
#define INFINITY 2147483647L             /* largest integer */
/* The interior of the image is easier (and safer) to use in calculating
   the tip estimate. At the edges part of the tip extends beyond the
   edge where the image values are unknown. It is possible to handle
   this case by making worst case assumptions and thereby extract some
   information even from the data near the edge. Below, I separate the code
   into two parts, one which handles the interior and one which handles
   the edges. This serves two functions. First, because the edges can
   be tricky to handle properly, I had more than the usual share of bugs
   in this part of the code during development. It was useful to be able
   to turn this part of the code on or off at will. Second: The execution
   of the routine is slightly faster with the code divided since it is not
   then necessary to do checking for special cases in the part of the code
   that handles the interior. 

*/

long itip_estimate_iter(image,im_xsiz,im_ysiz,tip_xsiz,tip_ysiz,
   xc,yc,tip0,thresh)
   long **image,         /* image array */
       im_xsiz,im_ysiz,           /* x,y size of image array */   
       tip_xsiz,tip_ysiz,         /* x,y, size of tip to be determined */
       xc,yc,           /* center coordinates of tip */
       **tip0,          /* starting est. Size must be tip_xsiz by tip_ysiz */
       thresh;          /* threshold value to use */
{
   long ixp,jxp;         /* index into the image (x') */
   long **open;  

   long count=0;         /* counts places where tip estimate is improved */

   open = iopen(image,im_xsiz,im_ysiz,tip0,tip_xsiz,tip_ysiz);

   for (jxp=tip_ysiz-1-yc;jxp<=im_ysiz-1-yc;jxp++) {
      for (ixp=tip_xsiz-1-xc;ixp<=im_xsiz-1-xc;ixp++) {
         if (image[jxp][ixp]-open[jxp][ixp] > thresh)
            if (itip_estimate_point(ixp,jxp,image,im_xsiz,im_ysiz,tip_xsiz,
               tip_ysiz,xc,yc,tip0,thresh)) count++;
      }
   }

   freematrix(open,im_ysiz);
   return(count);
}

/* 
   The following is a routine to do an initial estimate of the tip
   shape by using only a few selected points within the image. If the
   points are well-chosen this can produce most of the tip shape
   refinement with a small fraction of the compute time.
*/
void itip_estimate0(image,im_xsiz,im_ysiz,tip_xsiz,tip_ysiz,
   xc,yc,tip0,thresh)
   long **image,            /* image array */
       im_xsiz,im_ysiz,    /* x,y size of image array */   
       tip_xsiz,tip_ysiz,  /* x,y, size of tip to be determined */
       xc,yc,              /* center coordinates of the tip */
       **tip0,   /* starting estimate. Must be tip_xsiz by tip_ysiz in size */
       thresh;             /* threshold value to use */
{
   long i,j,n;
   long arraysize; /* size of array allocated to store list of image maxima */
   long *x,*y;	  /* point to coordinates of image maxima */
   long iter=0;
   long count;
   long delta;	/* defines what is meant by near neighborhood for purposes
		   of point selection. */
   /* 
      We need to create temporary arrays to hold a list of selected
      image coordinates.  The space needed depends upon the image. This
      creates a memory management issue. We address it by allowing for
      300 maxima--a good number (corresponding perhaps to a grainy
      surface) which should be big enough for most images. Then we
      monitor memory usage and reallocate a larger array if necessary.

      Coordinates to be used are determined by the routine useit, below,
      which returns either 1 (if the given coordinate is to be used) or
      0 (if not). The user can substitute his own version of this routine
      if he believes he has a more economical algorithm for choosing
      points.
   */
   arraysize = 300;
   x = (long *)malloc(arraysize*sizeof(long));
   if (x == NULL) {
      printf("Unable to allocate x array in itip_estimate0 routine.\n");
      return;
   }
   y = (long *)malloc(arraysize*sizeof(long));
   if (y == NULL) {
      printf("Unable to allocate y array in itip_estimate0 routine.\n");
      free(x);
      return;
   }

   /* 
   Now choose a nearest neighborhood size to send the useit routine.
   The neighborhood should be at least equal to 1 (i.e. consider all
   points with x,y within +/- 1 of the one under consideration).
   Otherwise ALL points are used, equivalent to the full tip_estimate
   routine, and no speed advantage is derived, which loses the whole
   point of having a tip_estimate0. However, the size of the
   neighborhood should in principle scale with the size of the tip. I
   use a small fraction of tip size (1/10) because in practice the
   routine seems to run acceptably quickly even at this
   setting--there's no point in sacrificing performance for speed if
   the present speed is acceptable.
   */

   delta = MAX(MAX(tip_xsiz,tip_ysiz)/10,1);

   /* Create a list of coordinates to use */
   n=0;		/* Number of image maxima found so far */
   for (j=tip_ysiz-1-yc;j<=im_ysiz-1-yc;j++) {
      for (i=tip_xsiz-1-xc;i<=im_xsiz-1-xc;i++) {
         if (useit(i,j,image,im_xsiz,im_ysiz,delta)) {
            if (n == arraysize) {   /* need more room in temporary arrays */
               arraysize *= 2;	/* increase array size by factor of 2 */
               x = (long *)realloc(x,arraysize*sizeof(long));
               if (x == NULL) {
                  printf("Unable to realloc x array in itip_estimate0.\n");
                  free(y);
                  return;
               }
               y = (long *)realloc(y,arraysize*sizeof(long));
               if (y == NULL) {
                  printf("Unable to realloc y array in itip_estimate0.\n");
                  free(x);
                  return;
               }
            }
            x[n] = i;	/* We found another one */
            y[n] = j;
            n++;
         }
      }
   }
   printf("Found %ld internal local maxima\n",n);
   

   /* Now refine tip at these coordinates recursively until no more change */
   do {
      count = 0;
      iter++;
      for (i=0;i<n;i++)
         if (itip_estimate_point(x[i],y[i],image,im_xsiz,im_ysiz,
            tip_xsiz,tip_ysiz,xc,yc,tip0,thresh)) count++;
      printf 
         ("Finished iteration #%ld. %ld image locations produced refinement.\n",
         iter,count);
   } while (count);

   /* free temporary space */
   free(x); free(y);
}

/* 
   The following is a routine that determines whether a selected point
   at coordintes x,y within an image is deemed to be suitable for
   image refinement. In this implementation, the algorithm simply
   decides to use the point if it is a local maximum of the image.
   It defines a local maximum as a point with height greater than any
   of its near neighbors, provided there are not too many near neighbors
   with values equal to the maximum (which indicates a flat).
*/

long useit(x,y,image,sx,sy,delta)
   long x,y;	/* Coordinates of selected point within image */
   long **image;	/* pointer to image */
   long sx,sy;	/* x and y size of image */
   long delta;	/* size of neighborhood to search for maximum (Searches
		   interval ([x-delta,x+delta],[y-delta,y+delta]) */
{
   long xmin,xmax,ymin,ymax;	/* actual interval to search */
   long i,j;
   long max=image[y][x]; /* value of maximum height in the neighborhood */
   long count=0;	/* counts # spots where pixel = max value */

   xmin = MAX(x-delta,0);
   xmax = MIN(x+delta,sx-1);
   ymin = MAX(y-delta,0);
   ymax = MIN(y+delta,sy-1);

   for (j=ymin;j<=ymax;j++) {
      for (i=xmin;i<=xmax;i++) {
         max = image[j][i]>=max?(count++,image[j][i]):max;
      }
   }   

   /* If the point equals the maximum value in the neighborhood we use it,
      unless there are too many points in the neighborhood with the same
      property--i.e. the neighborhood is flat */
   if (max == image[y][x] && count <= ((2*delta+1)^2)/5) return(1);
   return(0);
}

/* 
   The following routine does the same thing as itip_estimate_iter, except
   that instead of looping through all i,j contained within the image, it
   computes the tip shape as deduced from a single i,j coordinate. For what
   is this useful? The order of evaluation of the points can affect the
   execution speed. That is because the image at some locations puts great
   constraints on the tip shape. If the tip shape is refined by considering
   these points first, time is saved later. The following routine, since it
   performs the calculation at a single point, allows the user to select
   the order in which image coordinates are considered. In using the 
   routine in this mode, the fact that the refined tip replaces tip0 means
   results of one step automatically become the starting point for the next
   step. The routine returns an integer which is the number of pixels 
   within the starting tip estimate which were updated.

   To compile codes which does not use parts of the image within a tipsize
   of the edge, set USE_EDGES to 0 on the next line.   
*/
	
#define USE_EDGES 1			/* set to 1 to include edges */

long itip_estimate_point(ixp,jxp,image,im_xsiz,im_ysiz,tip_xsiz,tip_ysiz,
   xc,yc,tip0,thresh)
   long ixp,jxp,	/* image x,y coordinate to be considered */
       **image,         /* image array */
       im_xsiz,im_ysiz,           /* x,y size of image array */   
       tip_xsiz,tip_ysiz,         /* x,y, size of tip to be determined */
       xc,yc,           /* center coordinates of tip */
       **tip0,          /* starting est. Size must be tip_xsiz by tip_ysiz */
       thresh;          /* threshold value to use */
{
   long ix,jx,           /* index into the output tip array (x) */
       id,jd;           /* index into p' (d) */
   long temp,imagep,
       dil;   		 /* various intermediate results */
   long count=0;         /* counts places where tip estimate is improved */
   long interior;	/* =1 if no edge effects to worry about */
   int apexstate,
       xstate,
       inside = 1,	/* Point is inside image */
       outside = 0;	/* point is outside image */


   interior = jxp>=tip_ysiz-1 && jxp<=im_ysiz-tip_ysiz && ixp>=tip_xsiz-1 && 
             ixp<=im_xsiz-tip_xsiz;


   if (interior) {
      for (jx=0;jx<tip_ysiz;jx++) {
         for (ix=0;ix<tip_xsiz;ix++) {
	    /* First handle the large middle area where we don't have to
	       be concerned with edge problems. Because edges are far
	       away, we can leave out the overhead of checking for them
	       in this section. */
	       imagep = image[jxp][ixp];
	    dil = MINUS_INF;         /* initialize maximum to -infinity */
	    for (jd=0;jd<tip_ysiz;jd++) {
	       for (id=0;id<tip_xsiz;id++) {
		  if (imagep-image[jxp+yc-jd][ixp+xc-id] > 
		     tip0[jd][id]) continue;
		  temp = image[jx+jxp-jd][ix+ixp-id]+tip0[jd][id]-imagep;
		  dil = MAX(dil,temp);
	       } /* end for id */
	    } /* end for jd */
	    if (dil == MINUS_INF) continue;
	    tip0[jx][ix] = 
	       dil<tip0[jx][ix]-thresh?(count++,dil+thresh):tip0[jx][ix];
         } /* end for ix */
      } /* end for jx */
      return(count);
   } /* endif */
#if USE_EDGES
	    /* Now handle the edges */
   for (jx=0;jx<tip_ysiz;jx++) {
      for (ix=0;ix<tip_xsiz;ix++) {
	 imagep = image[jxp][ixp];
	 dil = MINUS_INF;         /* initialize maximum to -infinity */
	 for (jd=0;jd<=tip_ysiz-1 && dil<INFINITY;jd++) {
	    for (id=0;id<=tip_xsiz-1;id++) {
               /* Determine whether the tip apex at (xc,yc) lies within
                  the domain of the translated image, and if so, if it
                  is inside (i.e. below or on the surface of) the image. */
               apexstate = outside;	/* initialize */
               if (jxp+yc-jd<0 || jxp+yc-jd>=im_ysiz ||
                   ixp+xc-id<0 || ixp+xc-id>=im_xsiz) apexstate = inside;
               else if (imagep-image[jxp+yc-jd][ixp+xc-id] <= 
	             tip0[jd][id]) apexstate = inside;
               /* Determine whether the point (ix,jx) under consideration
                  lies within the domain of the translated image */
               if (jxp+jx-jd<0 || jxp+jx-jd>=im_ysiz ||
                   ixp+ix-id<0 || ixp+ix-id>=im_xsiz) xstate = outside;
               else xstate = inside;

               /* There are 3 actions we might take, depending upon
                  which of 4 states (2 apexstate possibilities times 2 
                  xstate ones) we are in. */

               /* If apex is outside and x is either in or out no change
                  is made for this (id,jd) */
               if (apexstate==outside) continue;

               /* If apex is inside and x is outside
		  worst case is translated image value -> INFINITY.
		  This would result in no change for ANY (id,jd). We
		  therefore abort the loop and go to next (ix,jx) value */
               if (xstate==outside) goto nextx;

               /* The only remaining possibility is x and apex both inside. 
                  This is the same case we treated in the interior. */
	       temp = image[jx+jxp-jd][ix+ixp-id]+tip0[jd][id]-imagep;
               dil = MAX(dil,temp);
	    } /* end for id */
	 } /* end for jd */
	 if (dil == MINUS_INF) continue;
	
	 tip0[jx][ix] = 
	    dil<tip0[jx][ix]-thresh?(count++,dil+thresh):tip0[jx][ix];
nextx:;
      } /* end for ix */
   } /* end for jx */
#endif

   return(count);
}


